class CsverveMergeDtypesEmptyMergeSet(Exception):
    pass


class CsverveMergeCommonColException(Exception):
    pass


class DtypesMergeException(Exception):
    pass


class CsverveConcatException(Exception):
    pass


class CsverveAnnotateError(Exception):
    pass


class CsverveMergeException(Exception):
    pass


class CsverveMergeColumnMismatchException(Exception):
    pass


class CsverveParseError(Exception):
    pass


class CsverveInputError(Exception):
    pass


class CsverveWriterError(Exception):
    pass


class CsverveDtypeError(Exception):
    pass
